#!/usr/bin/env python3
# encoding: utf-8
"""
Config Switcher.
"""
import argparse
import configparser
import copy
import os
import re
import shutil
import sys
from concurrent.futures import ThreadPoolExecutor
from time import sleep
from urllib.parse import urlparse

from csw.modules.helpers.config import get_config
from csw.modules.helpers.config import open_config_in_file_manager
from csw.modules.helpers.messages import print_error_and_exit
from csw.modules.helpers.messages import print_warning
from csw.modules.helpers.path import get_path
from csw.modules.helpers.ssh import download_remote_dir
from csw.modules.helpers.ssh import get_remote_file_content
from csw.modules.helpers.ssh import local_port_forwarding

SETTINGS = {}


def print_db_info(config_data):
    """
    Вывести информацию о подключении к базе.

    :param config_data: Данные из web_edu.conf
    :type config_data: dict
    """
    print('\nИнформация о базах данных:')
    print('DB_HOST: {}'.format(config_data['db_host']))
    print('DB_PORT: {}'.format(config_data['db_port']))
    print('DB_NAME: {}'.format(config_data['db_name']))
    print('DB_USER: {}'.format(config_data['db_user']))
    print('DB_PASSWORD: {}'.format(config_data['db_password']))

    print('SERVICE_DB_HOST: {}'.format(config_data['service_db_host']))
    print('SERVICE_DB_PORT: {}'.format(config_data['service_db_port']))
    print('SERVICE_DB_NAME: {}'.format(config_data['service_db_name']))


def print_server_info(addr, all_servers_info):
    """
    Вывести дополнительную информацию по серверу.

    :param addr: Адрес с портом
    :type addr: str
    :param all_servers_info: Доп информация по всем серверам
    :type all_servers_info: dict
    """
    server_info = all_servers_info.get(addr)  # type: dict
    if not server_info:
        return

    print('\nДоп. информация по серверу:')
    for key, value in server_info.items():
        print('{}: {}'.format(key, value))


def reset_configs():
    """
    Сброс конфигов на дефолтные.
    """
    files = (
        ('web_edu.conf.default', 'web_edu.conf'),
        ('plugins.conf.default', 'plugins.conf'),
    )

    for default_conf, conf in files:
        default_conf_path = get_path(
            SETTINGS['local_config_path'],
            default_conf,
        )
        conf_path = get_path(SETTINGS['local_config_path'], conf)

        try:
            shutil.copy2(default_conf_path, conf_path)
        except FileNotFoundError:
            print_warning(
                'Не найден конфиг по умолчанию {}'.format(default_conf)
            )


def load_remote_config(host, configs_path):
    """
    Получить текст конфигов с сервера.

    :param host: IP адрес сервера
    :type host: str
    :param configs_path: Путь до папки конфигов на сервере
    :type configs_path: str
    :return: Строки с содержимым web_edu.conf и plugins.conf
    :rtype: (str, str)
    """
    web_edu_conf = get_remote_file_content(
        host,
        os.path.join(configs_path, 'web_edu.conf'),
    )
    plugins_conf = get_remote_file_content(
        host,
        os.path.join(configs_path, 'plugins.conf'),
    )

    return web_edu_conf, plugins_conf


def load_local_config():
    """
    Получить текст локальных конфигов.

    :return: Строки с содержимым web_edu.conf и plugins.conf
    :rtype: (str, str)
    """
    with open(get_path(SETTINGS['local_config_path'], 'web_edu.conf'),
              encoding='utf-8') as web_edu_conf_file:
        web_edu_conf = web_edu_conf_file.read()

    with open(get_path(SETTINGS['local_config_path'], 'plugins.conf'),
              encoding='utf-8') as plugins_conf_file:
        plugins_conf = plugins_conf_file.read()

    return web_edu_conf, plugins_conf


def parse_web_conf(web_edu_conf, app_host):
    """
    Получить необходимую информацию из конфига настроек.

    :param web_edu_conf: Строка с содержимым файла web_edu.conf
    :type web_edu_conf: str
    :param app_host: Адрес хоста приложения
    :type app_host: str
    :return: Словарь с необходимыми данными
    :rtype: dict
    """
    config = configparser.ConfigParser(strict=False)
    config.read_string(web_edu_conf)

    extra_config_fields = copy.deepcopy(
        SETTINGS.get('extra_config_fields', [])
    )

    for section_name in extra_config_fields:
        for key in extra_config_fields[section_name]:
            default_value = extra_config_fields[section_name][key]

            if config.has_section(section_name):
                config_value = config[section_name].get(key, default_value)
            else:
                config_value = default_value

            if config_value is None:
                print_error_and_exit(
                    'Значение {} в конфиге не определено'.format(key)
                )

            extra_config_fields[section_name][key] = config_value

    def get_db_host(db_host):
        """
        Возвращает адрес хоста базы.

        :param db_host: Адрес хоста из конфига приложения
        :type db_host: str
        :return: Адрес хоста базы
        :rtype: str
        """
        if not db_host or db_host in ('localhost', '127.0.0.1'):
            return app_host

        return db_host

    config_data = {
        'db_host': get_db_host(config['database'].get('DATABASE_HOST')),
        'db_port': config['database'].get('DATABASE_PORT') or '5432',
        'db_name': config['database']['DATABASE_NAME'],
        'db_user': config['database']['DATABASE_USER'],
        'db_password': config['database']['DATABASE_PASSWORD'],

        'service_db_host': get_db_host(config['service_database'].get('HOST')),
        'service_db_port': config['service_database'].get('PORT') or '5432',
        'service_db_name': config['service_database']['NAME'],
        'service_db_user': config['service_database']['USER'],
        'service_db_password': config['service_database']['PASSWORD'],

        'extra_config_fields': extra_config_fields
    }

    if SETTINGS.get('port_forwarding'):
        config_data.update({
            'db_host': 'localhost',
            'db_port': str(SETTINGS['local_db_port']),
            'real_db_host': (config['database'].get('DATABASE_HOST')
                             or 'localhost'),
            'real_db_port': config_data['db_port'],

            'service_db_host': 'localhost',
            'service_db_port': str(SETTINGS['local_db_port']),
            'real_service_db_host': (config['service_database'].get('HOST')
                                     or 'localhost'),
            'real_service_db_port': config_data['service_db_port'],
        })

    return config_data


def set_web_conf(config_data):
    """
    Подставить данные в конфиг настроек.

    :param config_data: Данные из web_edu.conf
    :type config_data: dict
    """
    config_file_path = get_path(SETTINGS['local_config_path'], 'web_edu.conf')

    config = configparser.ConfigParser()
    config.read(config_file_path, encoding='utf-8')

    def _create_config_section(section):
        if section not in config:
            config[section] = {}

    # Установка флага использования CSW
    _create_config_section('dev')
    config['dev']['CSW_USED'] = 'True'

    config['database']['DATABASE_HOST'] = config_data['db_host']
    config['database']['DATABASE_PORT'] = config_data['db_port']
    config['database']['DATABASE_NAME'] = config_data['db_name']
    config['database']['DATABASE_USER'] = config_data['db_user']
    config['database']['DATABASE_PASSWORD'] = config_data['db_password']

    config['service_database']['HOST'] = config_data['service_db_host']
    config['service_database']['PORT'] = config_data['service_db_port']
    config['service_database']['NAME'] = config_data['service_db_name']
    config['service_database']['USER'] = config_data['service_db_user']
    config['service_database']['PASSWORD'] = config_data['service_db_password']

    extra_config_fields = config_data['extra_config_fields']
    for section_name in extra_config_fields:
        _create_config_section(section_name)
        config[section_name].update(extra_config_fields[section_name])

    with open(config_file_path, 'w', encoding='utf-8') as config_file:
        config.write(config_file)


def set_plugin_conf(config_data, disabled_plugins):
    """
    Подставить данные в конфиг плагинов.

    :param config_data: Строка с содержимым файла plugins.conf
    :type config_data: str
    :param disabled_plugins: Список отключенных плагинов
    :type disabled_plugins: list of str
    """
    if not disabled_plugins:
        disabled_plugins = []

    config_file_path = get_path(SETTINGS['local_config_path'], 'plugins.conf')

    config = configparser.ConfigParser()
    config.read_string(config_data)

    plugins_list = config['plugins']['activated_plugins'].replace(
        '\n', ''
    ).split(',')
    plugins_list = [plugin for plugin in plugins_list
                    if plugin not in disabled_plugins]
    config['plugins']['activated_plugins'] = ',\n'.join(plugins_list)

    with open(config_file_path, 'w', encoding='utf-8') as config_file:
        config.write(config_file)


def download_conf_folder(host, config_path):
    """
    Загрузка папки с конфигами с удаленного сервера.

    :param host: IP адрес сервера
    :type host: str
    :param config_path: Путь до папки конфигов на сервере
    :type config_path: str
    """
    download_conf_path = SETTINGS.get('download_conf_path')
    if not download_conf_path:
        return

    download_conf_path = get_path(download_conf_path)
    file_extensions = (
        '.conf',
        '.config',
        '.yaml',
        '.xml',
        '.json',
        '.py',
        '.crt',
        '.key',
        '.pem',
    )

    # TODO: Сделать какой-нибудь индикатор загрузки
    ThreadPoolExecutor().submit(  # pylint: disable=consider-using-with
        download_remote_dir,
        host,
        config_path,
        download_conf_path,
        file_extensions
    )


def parse_addr(addr):
    """
    Разбор строки с адресом.

    :param addr: Строка с адресом приложения
    :type addr: str
    :return: Полный адрес, хост, порт
    :rtype: (str, str, str)
    """
    # Приведем host:port и http://host:port/ к одному виду
    addr = urlparse(addr).netloc or urlparse(addr).path

    # Разделим host и port с учетом, что port может не быть
    pattern = re.compile(r'([\w.-]+)(:([0-9]+))?')
    match = pattern.match(addr)
    try:
        host = match.group(1)
        port = match.group(3)
    except (IndexError, AttributeError):
        print_error_and_exit('Неверно указан адрес')
        sys.exit(-1)

    return addr, host, port


def resolver(addr):
    """
    Преобразование адресов.

    :param addr: Строка с адресом приложения
    :type addr: str
    :return: Полный адрес, хост, порт
    :rtype: (str, str, str)
    """
    addr, host, port = parse_addr(addr)

    while True:
        addr = '{}:{}'.format(host, port) if port else host

        # Если можно получить настройки, значит адрес правильный
        if SETTINGS['config_path'].get(addr):
            break

        host_port = (SETTINGS['resolver'].get(addr)
                     or SETTINGS['resolver'].get(host))
        if not host_port:
            break

        try:
            host, port = host_port.split(':')
        except ValueError:
            host = host_port

    return addr, host, port


def switch_configs(addr):
    """
    Подмена конфига.

    :param addr: Строка с адресом приложения
    :type addr: str
    """
    # Преобразуем имена в IP и внешние ip во внутренние
    addr, host, _ = resolver(addr)

    config_path = SETTINGS['config_path'].get(addr)
    if not config_path:
        print_error_and_exit('Адрес не найден в списке известных')

    reset_configs()

    r_settings_conf, r_plugins_conf = load_remote_config(host, config_path)

    web_conf = parse_web_conf(r_settings_conf, host)
    set_web_conf(web_conf)

    set_plugin_conf(r_plugins_conf, SETTINGS['disabled_plugins'])

    print_db_info(web_conf)
    print_server_info(addr, SETTINGS['server_info'])

    download_conf_folder(host, config_path)

    print('\nПосле окончания работ с тестовым сервером используйте '
          'комбинацию Ctrl+C для завершения работы утилиты.')
    try:
        if SETTINGS.get('port_forwarding'):
            print('\nПробрасываем порт.')
            local_port_forwarding(
                host,
                SETTINGS['local_db_port'],
                web_conf['real_db_host'],
                web_conf['real_db_port']
            )
        else:
            while True:
                sleep(1)
    except KeyboardInterrupt:
        pass
    finally:
        print('\nВозвращаем конфиги к дефолтным')
        reset_configs()


def parse_args():
    """
    Разбор аргументов командной строки.

    :return: Разобранные аргументы
    :rtype: argparse.Namespace
    """
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    exclusive_group = parser.add_mutually_exclusive_group(required=True)

    exclusive_group.add_argument(
        'addr',
        nargs='?',
        help='Адрес приложения, к которому производится подключение',
    )
    parser.add_argument(
        '-f', '--port-forward',
        action='store_true',
        help='Пробрасывать порт к базе через сервер приложения',
    )

    exclusive_group.add_argument(
        '-c', '--config',
        action='store_true',
        help='Открыть папку с настройками скрипта',
    )
    exclusive_group.add_argument(
        '-r', '--reset',
        action='store_true',
        help='Сбросить файлы конфигурации до конфигурации по умолчанию',
    )

    args = parser.parse_args()

    return args


def main():
    """
    Точка входа.
    """
    # pylint: disable=global-statement
    global SETTINGS

    args = parse_args()

    if args.config:
        open_config_in_file_manager()
        return

    SETTINGS = get_config()

    if args.reset:
        print('Сброс')
        reset_configs()
        return

    if args.port_forward:
        SETTINGS['port_forwarding'] = True

    switch_configs(sys.argv[1])


if __name__ == '__main__':
    main()
