#!/usr/bin/env python3
# encoding: utf-8
"""
Dump/Restore.
"""
import argparse
import json
import os
import re
import subprocess  # noqa: S404
from collections import namedtuple
from datetime import datetime
from pathlib import Path

import csw
from csw.app import parse_web_conf
from csw.app import resolver
from csw.modules.helpers.config import get_config
from csw.modules.helpers.messages import print_error_and_exit
from csw.modules.helpers.messages import print_warning
from csw.modules.helpers.path import get_path
from csw.modules.helpers.ssh import get_remote_file_content

SETTINGS = {}

DBConnectParameters = namedtuple(
    'DBConnectParameters',
    [
        'host',
        'port',
        'db_name',
        'username',
        'password',
    ]
)


class DBServer:
    """
    Подключение к базе данных.
    """

    def __init__(self, host='localhost', port='5432',
                 db_name='postgres', db_user='postgres'):
        """
        Инициализация.

        :param host: Хост
        :type host: str
        :param port: Порт
        :type port: str | int
        :param db_name: Имя базы
        :type db_name: str
        :param db_user: Имя пользователя
        :type db_user: str
        """
        self.host = host
        self.port = str(port)
        self.db_name = db_name
        self.db_user = db_user

    def __eq__(self, other):
        """
        Переопределение функции сравнения.

        Два экземпляра будут равны при равных параметрах подключения

        :param other: Сравниваемый объект
        :type other: Any
        :return: Результат
        :rtype: bool
        """
        if not isinstance(other, DBServer):
            return NotImplemented

        return self.__dict__ == other.__dict__

    def __str__(self):  # noqa: D105
        return (
            '{class_name}: {host}:{port}/{db_name}'.format(
                class_name=self.__class__.__name__,
                host=self.host,
                port=self.port,
                db_name=self.db_name,
            )
        )

    def run_psql_command(self, command, return_value=False, **kwargs):
        """
        Выполнить psql команду.

        :param command: SQL команда
        :type command: str
        :param return_value: Возвращать значение
        :type return_value: bool
        :param kwargs: Параметры выполнения команды
        :type kwargs: str
        :return: Результат выполнения команды
        :rtype: str | None
        """
        command_args = [
            'psql',
            '--host',
            kwargs.get('host') or self.host,
            '--port',
            kwargs.get('port') or self.port,
            '--dbname',
            kwargs.get('db_name') or self.db_name,
            '--username',
            kwargs.get('db_user') or self.db_user,
            '--no-password',
            '--tuples-only',
            '--command',
            command,
        ]

        try:
            result = subprocess.run(  # noqa: S603
                command_args,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=True,
            )
        except subprocess.CalledProcessError as exc:
            err_msg = exc.stderr.decode()
            print_error_and_exit(
                'Ошибка выполнения команды psql:\n{}'.format(err_msg)
            )

        if return_value:
            return result.stdout.decode().strip()

        return None

    def get_db_size(self):
        """
        Получить размер базы.

        :return: Строка с размером
        :rtype: str
        """
        sql_command = (
            "select pg_size_pretty(pg_database_size('{}'));"
        ).format(self.db_name)

        result = self.run_psql_command(sql_command, return_value=True)

        return result

    def get_table_size(self, table_name):
        """
        Получить размер таблицы.

        :param table_name: Имя таблицы
        :type table_name: str
        :return: Строка с размером
        :rtype: str
        """
        sql_command = (
            "select pg_size_pretty(pg_total_relation_size('{}'));"
        ).format(table_name)

        result = self.run_psql_command(sql_command, return_value=True)

        return result

    def get_db_users(self):
        """
        Получить пользователей, которым даны права к базе.

        :return: Множество пользователей
        :rtype: set
        """
        # Пользователи, привязанные к самой базе
        db_users_sql = """
            select
              json_agg(split_part(datacl, '='::text, 1))
            from (
              select
                unnest(datacl)::text datacl
              from
                pg_database
              where
                datname = '{}'
            ) x
        """.format(self.db_name)  # noqa: S608

        db_users_raw = self.run_psql_command(db_users_sql, return_value=True)

        db_users = json.loads(db_users_raw) if db_users_raw else []
        db_users = [user for user in db_users if user]

        # Пользователи, привязанные к таблицам базы
        tables_users_sql = (
            'select json_agg(distinct grantee) '
            'from information_schema.role_table_grants '
        )
        tables_users_raw = self.run_psql_command(
            tables_users_sql,
            return_value=True,
        )

        tables_users = json.loads(tables_users_raw) if tables_users_raw else []
        tables_users = [user
                        for user in tables_users
                        if user and user != 'PUBLIC']

        users = set(db_users + tables_users)

        return users

    def copy_db_privileges(self, target_db):
        """
        Копировать доступы самой базы.

        :param target_db: Целевая база
        :type target_db: DBServer
        """
        print(
            'Получение информации о пользователях базы {}'.format(self.db_name)
        )

        remote_db_users_list = self.get_db_users()

        if not remote_db_users_list:
            return

        local_users_sql = 'select json_agg(rolname) from pg_roles'
        local_users_raw = target_db.run_psql_command(
            local_users_sql,
            db_name='postgres',
            return_value=True,
        )
        local_users = json.loads(local_users_raw)

        new_users = set(remote_db_users_list) - set(local_users)

        if not new_users:
            return

        command_args = [
            'pg_dumpall',
            '--host',
            self.host,
            '--port',
            self.port,
            '--database',
            self.db_name,
            '--username',
            self.db_user,
            '--no-password',
            '--roles-only',
        ]

        try:
            result = subprocess.run(  # noqa: S603
                command_args,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=True
            )
        except subprocess.CalledProcessError as exc:
            err_msg = exc.stderr.decode()
            print_error_and_exit(
                'Ошибка выполнения команды pg_dumpall:\n{}'.format(err_msg)
            )

        roles_raw = result.stdout.decode()

        commands = []
        for line in roles_raw.splitlines():
            if not any(user in line for user in new_users):
                continue

            # Уберем информацию о пароле и сроке действия
            line = re.sub("password '[^']+'", '', line, flags=re.IGNORECASE)
            line = re.sub("valid until '[^']+'", '', line, flags=re.IGNORECASE)

            commands.append(line)

        for command in commands:
            target_db.run_psql_command(command)

    def dump(self, save_path, tables_list=None, excluded_tables_list=None):
        """
        Сохранить дамп базы.

        :param save_path: Папка сохранения дампа
        :type save_path: str
        :param tables_list: Список таблиц для дампа
        :type tables_list: list
        :param excluded_tables_list: Список таблиц,
                                     данные из которых не будут переноситься
        :type excluded_tables_list: list
        :return: Путь до файла с дампом
        :rtype: str
        """
        backup_path = os.path.join(
            save_path,
            '{}_{}.backup'.format(
                self.db_name,
                datetime.now().strftime('%Y-%m-%d_%H:%M'),
            )
        )

        if not tables_list:
            tables_list = []

        if not excluded_tables_list:
            excluded_tables_list = []

        command_args = [
            'pg_dump',
            '--file',
            backup_path,
            '--host',
            self.host,
            '--port',
            self.port,
            '--username',
            self.db_user,
            '--no-password',
            '--verbose',
            '--format=custom',
            '--blobs',
            *['--table={}'.format(table) for table in tables_list],

            # Внутри шаблона * обозначает любое количество символов,
            # включая отсутствие символов.
            # ? соответствует любому одному символу
            *['--exclude-table-data={}'.format(table)
              for table in excluded_tables_list],
            self.db_name,
        ]

        print('Получение дампа базы {}'.format(self.db_name))
        try:
            subprocess.run(  # noqa: S603
                command_args,
                # TODO: Разобраться с выводом
                # stdout=subprocess.STDOUT,  # Вывод в консоль
                # stderr=subprocess.PIPE,
                check=True,
            )
        except subprocess.CalledProcessError as exc:
            err_msg = exc.stderr.decode()
            print_error_and_exit(
                'Ошибка выполнения команды pg_dump:\n{}'.format(err_msg)
            )

        return backup_path

    def get_last_dump(self, save_path):
        """
        Получить путь до последнего сохраненного дампа базы.

        :param save_path: Папка сохранения дампа
        :type save_path: str
        :return: Путь до файла с дампом
        :rtype: str
        """
        backup_name_pattern = re.compile(
            self.db_name + r'_\d{4}-\d{2}-\d{2}_\d{2}:\d{2}.backup'
        )

        last_dump_filename = max(f for f in os.listdir(save_path)
                                 if backup_name_pattern.match(f))

        return os.path.join(save_path, last_dump_filename)

    def reset_schema(self, schema_name):
        """
        Сбросить схему (удалить и создать заново).

        :param schema_name: Имя схемы
        :type schema_name: str
        """
        if self.host != 'localhost':
            print_error_and_exit('Удалять схему можно только на localhost')

        if self.db_name == 'postgres':
            print_error_and_exit('Попытка удаления схемы в базе postgres')

        sql_command = (
            'drop schema if exists {0} cascade; '
            'create schema {0}; '
        ).format(schema_name)

        self.run_psql_command(sql_command)

    def close_connections(self):
        """
        Завершить все подключения к базе.
        """
        if self.host != 'localhost':
            print_error_and_exit(
                'Сбрасывать подключения к базе можно только на localhost'
            )

        sql_command = (
            'select pg_terminate_backend(pid) '  # noqa: S608
            'from pg_stat_activity '
            "where datname = '{}';"
        ).format(self.db_name)

        self.run_psql_command(sql_command, db_name='postgres')

    def drop_db(self):
        """
        Удалить базу.
        """
        if self.host != 'localhost':
            print_error_and_exit('Удалять базу можно только на localhost')

        if self.db_name == 'postgres':
            print_error_and_exit('Попытка удаления базы postgres')

        self.close_connections()

        command_args = [
            'dropdb',
            '--host',
            self.host,
            '--port',
            self.port,
            '--username',
            self.db_user,
            '--no-password',
            '--if-exists',
            self.db_name,
        ]

        print('Удаление старой базы {}'.format(self.db_name))
        try:
            subprocess.run(  # noqa: S603
                command_args,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=True,
            )
        except subprocess.CalledProcessError as exc:
            err_msg = exc.stderr.decode()
            print_error_and_exit(
                'Ошибка выполнения команды dropdb:\n{}'.format(err_msg)
            )

    def restore(self, backup_path, tables_list=None):
        """
        Восстановить базу из дампа.

        :param backup_path: Путь до дампа
        :type backup_path: str
        :param tables_list: Список таблиц для восстановления
        :type tables_list: list
        """
        if self.host != 'localhost':
            print_error_and_exit(
                'Восстанавливать базу можно только на localhost'
            )

        if self.db_name == 'postgres':
            print_error_and_exit('Попытка восстановления базы postgres')

        self.drop_db()

        if not tables_list:
            tables_list = []

        command_args = [
            'pg_restore',
            '--host',
            self.host,
            '--port',
            self.port,
            '--username',
            self.db_user,
            '--no-password',
            '--create',
            '--clean',
            '--if-exists',
            '--dbname',
            'postgres',
            *['--table={}'.format(table) for table in tables_list],
            backup_path,
        ]

        print('Восстановление из дампа базы {}'.format(self.db_name))
        try:
            subprocess.run(  # noqa: S603
                command_args,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=True,
            )
        except subprocess.CalledProcessError as exc:
            err_msg = exc.stderr.decode()
            print_error_and_exit(
                'Ошибка выполнения команды pg_restore:\n{}'.format(err_msg)
            )


def parse_args():
    """
    Разбор аргументов командной строки.

    :return: Разобранные аргументы
    :rtype: argparse.Namespace
    """
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    exclusive_group = parser.add_mutually_exclusive_group(required=True)

    exclusive_group.add_argument(
        'addr',
        nargs='?',
        help='Адрес приложения, базу которого будем сохранять',
    )

    exclusive_group.add_argument(
        '-r', '--restore',
        help='Восстановить дамп из файла',
    )

    parser.add_argument(
        '-d', '--dump',
        action='store_true',
        help='Только сохранить дамп, без восстановления',
    )
    parser.add_argument(
        '-t', '--table',
        action='append',
        help=('Дамп/восстановление определенной таблицы '
              '(для нескольких таблиц указать аргумент несколько раз, '
              'например: -t table1 -t table2)'),
    )

    parser.add_argument(
        '-l', '--restore-last',
        action='store_true',
        help='Восстановить последний дамп приложения',
    )

    args = parser.parse_args()

    if args.restore_last and not args.addr:
        parser.error('При использовании флага --restore-last '
                     'необходимо указать адрес приложения')

    if args.addr and args.restore:
        parser.error('addr и --restore несовместимы')

    if args.dump and args.restore:
        parser.error('--dump и --restore несовместимы')

    return args.addr, args.dump, args.restore_last, args.restore, args.table


def get_db_parameters(addr):
    """
    Получить параметры подключения к базам.

    :param addr: Строка с адресом приложения
    :type addr: str
    :return: Параметры подключения к основной и сервисной базам
    :rtype: (DBConnectParameters, DBConnectParameters)
    """
    # FIXME: Костыль
    csw.app.SETTINGS = SETTINGS
    addr, host, _ = resolver(addr)

    # Путь до конфига на сервере
    config_path = SETTINGS['config_path'].get(addr)
    if not config_path:
        print_error_and_exit('Адрес не найден в списке известных')

    web_edu_conf = get_remote_file_content(
        host,
        os.path.join(config_path, 'web_edu.conf'),
    )

    parsed_conf = parse_web_conf(web_edu_conf, host)

    main_db = DBConnectParameters(
        parsed_conf['db_host'],
        parsed_conf['db_port'],
        parsed_conf['db_name'],
        parsed_conf['db_user'],
        parsed_conf['db_password'],
    )

    service_db = DBConnectParameters(
        parsed_conf['service_db_host'],
        parsed_conf['service_db_port'],
        parsed_conf['service_db_name'],
        parsed_conf['service_db_user'],
        parsed_conf['service_db_password'],
    )

    return main_db, service_db


def update_pgpass(db_param):
    """
    Проверка наличия в файле .pgpass данных для подключения.

    :param db_param: Параметры подключения к базе
    :type db_param: DBConnectParameters
    """
    pgpass_path = Path('~/.pgpass').expanduser()

    if not pgpass_path.exists():
        pgpass_path.touch(mode=0o600)

    with open(pgpass_path) as pgpass_file:
        pgpass = pgpass_file.read()

    pgpass_full_line = ':'.join(db_param)
    pgpass_db_name = ':'.join(db_param[:3]) + ':'

    # Есть полное совпадение
    if re.search('^' + pgpass_full_line, pgpass, re.MULTILINE):
        return

    # Есть совпадение до имени базы
    if re.search('^' + pgpass_db_name, pgpass, re.MULTILINE):
        pgpass = re.sub(
            pgpass_db_name + '.*',
            pgpass_full_line,
            pgpass,
            re.MULTILINE
        )
    else:
        pgpass = '{}{}{}\n'.format(
            pgpass,
            '\n' if not pgpass.endswith('\n') else '',
            pgpass_full_line,
        )

    with open(pgpass_path, 'w') as pgpass_file:
        pgpass_file.write(pgpass)


def restore_dump_from_file(dump_name, backup_path, local_db_port):
    """
    Восстановить дамп по имени файла.

    :param dump_name: Имя файла с дампом
    :type dump_name: str
    :param backup_path: Директория хранения дампов
    :type backup_path: str
    :param local_db_port: Порт для подключения к локальной базе
    :type local_db_port: int
    """
    # Передать можно, как полный путь, так и имя файла в папке с дампами
    if os.path.exists(dump_name):
        dump_dir_path, dump_name = os.path.split(dump_name)
    else:
        dump_dir_path = backup_path

    # Имя файла должно соответствовать шаблону,
    # например bars_web_edu_release_2020-07-15_23:57.backup
    dump_name_pattern = re.compile(
        r'(.+)_\d{4}-\d{2}-\d{2}_\d{2}:\d{2}.backup'
    )
    match = dump_name_pattern.match(dump_name)
    if not match:
        print_error_and_exit(
            '{}\nИмя дампа не соответствует шаблону'.format(dump_name)
        )

    db_name = match.group(1)

    dump_path = get_path(dump_dir_path, dump_name)
    if not os.path.exists(dump_path) or os.path.isdir(dump_path):
        print_error_and_exit(
            '{}\nДамп не найден'.format(dump_path)
        )

    local_db = DBServer(port=local_db_port, db_name=db_name)
    local_db.restore(dump_path)


def check_size(main_from, serv_from, tables_list):
    """
    Вывод размеров баз (и таблиц) и справшиваем о необходимости продолжения.

    :param main_from: Основная база, с которой будет сниматься дамп
    :type main_from: DBServer
    :param serv_from: Сервисная база, с которой будет сниматься дамп
    :type serv_from: DBServer
    :param tables_list: Список таблиц для дампа
    :type tables_list: list of str
    :return: Ответ пользователя на вопрос о продолжение работы
    :rtype: bool
    """
    print('Размер основной базы ({}): {}'.format(
        main_from.db_name,
        main_from.get_db_size(),
    ))
    print('Размер сервисной базы ({}): {}'.format(
        serv_from.db_name,
        serv_from.get_db_size()
    ))
    if tables_list:
        print('Размер выбранных таблиц:')
        for table in tables_list:
            try:
                table_size = main_from.get_table_size(table)
                print('\t{} - {}'.format(table, table_size))
            except RuntimeError:
                pass

    inp = input('Продолжить? [yN] ')  # noqa: S322
    if inp.lower() not in ('y', 'yes'):
        return False

    return True


def main():  # pylint: disable=too-many-locals
    """
    Точка входа.
    """
    # pylint: disable=global-statement
    global SETTINGS

    (addr, only_dump, restore_last,
     restore_dump_name, tables_list) = parse_args()

    SETTINGS = get_config()

    if not SETTINGS.get('dbdump'):
        print_warning(
            'В конфигурации отсутствует блок dbdump. '
            'Будут использованы параметры по умолчанию'
        )

    backup_path = get_path(
        SETTINGS.get('dbdump', {}).get('backup_path')
        or '/tmp/db_dumps/'  # noqa: S108
    )
    local_db_port = SETTINGS.get('dbdump', {}).get('local_db_port') or 5432
    excluded_tables = SETTINGS.get('dbdump', {}).get('excluded_tables', [])

    # Восстановление по имени файла
    if restore_dump_name:
        restore_dump_from_file(restore_dump_name, backup_path, local_db_port)
        return

    main_db_param, serv_db_param = get_db_parameters(addr)
    update_pgpass(main_db_param)
    update_pgpass(serv_db_param)

    main_from = DBServer(main_db_param.host, main_db_param.port,
                         main_db_param.db_name, main_db_param.username)
    serv_from = DBServer(serv_db_param.host, serv_db_param.port,
                         serv_db_param.db_name, serv_db_param.username)

    main_to = DBServer(port=local_db_port, db_name=main_db_param.db_name)
    serv_to = DBServer(port=local_db_port, db_name=serv_db_param.db_name)

    # Восстановление последнего дампа приложения
    if restore_last:
        main_dump = main_from.get_last_dump(backup_path)
        serv_dump = serv_from.get_last_dump(backup_path)
    else:
        if not check_size(main_from, serv_from, tables_list):
            return

        if not os.path.exists(backup_path):
            os.makedirs(backup_path)

        main_dump = main_from.dump(backup_path, tables_list, excluded_tables)
        serv_dump = serv_from.dump(backup_path, tables_list, excluded_tables)

        if only_dump:
            return

    main_from.copy_db_privileges(main_to)
    serv_from.copy_db_privileges(serv_to)

    main_to.restore(main_dump, tables_list)
    serv_to.restore(serv_dump, tables_list)


if __name__ == '__main__':
    main()
