# encoding: utf-8
"""
Вспомогательные функции. SSH.
"""
import os
import platform
import socket
import subprocess  # noqa: S404
from contextlib import contextmanager
from stat import S_ISDIR

import paramiko

from .config import get_config
from .messages import print_error_and_exit
from .path import get_path
from .path import remove_dir_content


def get_ssh_command():
    """
    Получить команду для запуска SSH.

    :return: Команда вызова SSH
    :rtype: str
    """
    # На не Windows можно вызвать просто ssh
    if platform.system() != 'Windows':
        return 'ssh'

    # На Windows приходится идти в обход
    # https://stackoverflow.com/a/41631476
    system32_dir = (
        'SysNative'
        if platform.architecture()[0] == '32bit'
        else 'System32'
    )

    ssh_path = os.path.join(
        os.environ['SystemRoot'],
        system32_dir,
        'OpenSSH'
    )

    if not os.path.exists(ssh_path):
        print_error_and_exit('OpenSSH не установлен в системе')

    return os.path.join(ssh_path, 'ssh.exe')


def get_ssh_config_and_key(host):
    """
    Получить конфиг SSH и ключ.

    :param host: Хост подключения
    :type host: str
    :return: Конфиг SSH и объект ключа
    :rtype: (dict, paramiko.RSAKey)
    """
    conf = get_config()

    ssh_config_path = get_path(conf['ssh_config_path'])
    private_key_path = get_path(conf['ssh_private_key_path'])

    with open(ssh_config_path, encoding='utf-8') as config_file:
        ssh_config = paramiko.config.SSHConfig()
        ssh_config.parse(config_file)

    host_config = ssh_config.lookup(host)
    if not host_config.get('user'):
        print_error_and_exit(
            'Не найдены данные для ssh соединения к хосту {}'.format(host)
        )

    key = paramiko.RSAKey.from_private_key_file(private_key_path)

    return host_config, key


@contextmanager
def ssh_client(host):
    """
    Контекстный менеджер для создания ssh-подключения.

    :param host: Хост подключения
    :type host: str
    :yield: Экземпляр класса подключения
    :rtype: paramiko.SSHClient
    """
    host_config, key = get_ssh_config_and_key(host)

    # TODO: Не создавать подключение каждый раз заново
    with paramiko.SSHClient() as client:
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        try:
            client.connect(
                host_config['hostname'],
                username=host_config['user'],
                pkey=key,
                timeout=3,
            )
            yield client

        except socket.timeout:
            print_error_and_exit(
                'Таймаут соединения к серверу.\n'
                'Сервер не отвечает или публичный ключ не загружен на сервер'
            )


def get_remote_file_content(host, file_path):
    """
    Получить содержимое файла с удаленного сервера.

    :param host: Хост подключения
    :type host: str
    :param file_path: Путь до файла на удаленном сервере
    :type file_path: str
    :return: Текст файла
    :rtype: str
    """
    # TODO: Переделать на что-то нормальное
    with ssh_client(host) as client:
        _, stdout, stderr = client.exec_command(  # noqa: S601
            'cat {}'.format(file_path)
        )

        err_msg = stderr.read().decode()
        if err_msg:
            print_error_and_exit(
                'Ошибка получения файла конфигурации '
                'с удаленного сервера:\n{}'.format(err_msg)
            )

        file_content = stdout.read().decode()

    return file_content


def local_port_forwarding(host, local_port, remote_host, remote_port,
                          only_print=False):
    """
    Локальный проброс порта.

    :param host: Хост
    :type host: str
    :param local_port: Локальный порт
    :type local_port: int | str
    :param remote_host: Удаленный хост
    :type remote_host: str
    :param remote_port: Удаленный порт
    :type remote_port: int | str
    :param only_print: Только вывести команду, без проброса
    :type only_print: bool
    """
    command_args = [
        get_ssh_command(),
        '-N',  # без консоли
        '-L',  # локальный проброс порта
        '{}:{}:{}'.format(local_port, remote_host, remote_port),
        host,
    ]

    if only_print:
        print(' '.join(command_args))
        return

    subprocess.run(command_args, check=True)  # noqa: S603


def sftp_walk(sftp, remote_path):
    """
    Рекурсивное получение имен файлов с полными путями.

    :param sftp: SFTPClient
    :type sftp: paramiko.SFTPClient
    :param remote_path: Путь к папке на сервере
    :type remote_path: str
    :return: Путь до папки и список файлов в папке
    :rtype: (str, list of str)

    # noqa: DAR301, DAR202
    """
    files = []
    folders = []

    for file_info in sftp.listdir_attr(remote_path):
        # Есть тяжелые папки 'media', в которых нет конфигов
        if file_info.filename == 'media':
            continue

        if S_ISDIR(file_info.st_mode):
            folders.append(file_info.filename)
        else:
            files.append(file_info.filename)

    if files:
        yield remote_path, files

    for folder in folders:
        folder_path = os.path.join(remote_path, folder)
        for folder_path, folder_files in sftp_walk(sftp, folder_path):
            yield folder_path, folder_files


def download_remote_dir(host, remote_path, local_path,
                        file_extensions=(), file_size_limit=1024):
    """
    Загрузить папку с удаленного сервера.

    :param host: Хост
    :type host: str
    :param remote_path: Путь к папке на сервере
    :type remote_path: str
    :param local_path: Локальная директория для загрузки директории
    :type local_path: str
    :param file_extensions: Расширения файлов, которые будут скачиваться
    :type file_extensions: tuple
    :param file_size_limit: Ограничение на размер 1 файла, в килобайтах
    :type file_size_limit: int
    """
    file_extensions = tuple(file_extensions)

    with ssh_client(host) as client:
        sftp = client.open_sftp()  # type: paramiko.SFTPClient

        if os.path.exists(local_path):
            remove_dir_content(local_path)

        for r_path, r_files in sftp_walk(sftp, remote_path):
            l_path = get_path(local_path, os.path.relpath(r_path, remote_path))
            for r_file in r_files:

                if file_extensions and not r_file.endswith(file_extensions):
                    continue

                # Размер файла в байтах
                file_size = sftp.stat(get_path(r_path, r_file)).st_size
                if file_size > file_size_limit * 1024:
                    continue

                if not os.path.exists(l_path):
                    os.makedirs(l_path, exist_ok=True)

                sftp.get(
                    get_path(r_path, r_file),
                    get_path(l_path, r_file),
                )
